/*
 * Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.dsl.processor.generator;

import javax.lang.model.element.*;

import com.oracle.truffle.dsl.processor.*;
import com.oracle.truffle.dsl.processor.java.model.*;

public abstract class AbstractCodeElementFactory<M> {

    protected final ProcessorContext context;
    private M model;

    private CodeElement<? super Element> element;

    public AbstractCodeElementFactory() {
        this.context = ProcessorContext.getInstance();
    }

    protected abstract CodeElement<?> create(M m);

    @SuppressWarnings("unused")
    protected void createChildren(M m) {
    }

    @SuppressWarnings({"unchecked", "rawtypes"})
    public CodeElement<?> process(CodeElement parent, M m) {
        model = m;
        element = (CodeElement<? super Element>) create(model);
        if (parent != null) {
            parent.add(element);
        }
        if (element != null) {
            createChildren(model);
        }
        return element;
    }

    @SuppressWarnings("rawtypes")
    public CodeElement getElement() {
        return element;
    }

    protected <MO, K extends Element> void add(AbstractCodeElementFactory<MO> factory, MO m) {
        factory.process(this.element, m);
    }

    public ProcessorContext getContext() {
        return context;
    }

    public M getModel() {
        return model;
    }

}
